@@ -1,13 +1,19 @@
 from __future__ import absolute_import, unicode_literals
 
+import inspect
 import optparse
 import os
 import shutil
+import subprocess
 import sys
 import tempfile
+import textwrap
 import zipfile
 
+import pypiserver
 import pytest
+import pytest_localserver.http
+import six
 
 import virtualenv
 
@@ -424,3 +430,125 @@ def test_create_environment_in_dir_with_spaces(tmpdir):
     """Should work with environment path containing spaces."""
     ve_path = str(tmpdir / "venv with spaces")
     virtualenv.create_environment(ve_path)
+
+
+def test_create_environment_with_local_https_pypi(tmpdir):
+    """Create virtual environment using local PyPI listening https with
+    certificate signed with custom certificate authority
+    """
+    test_dir = Path(__file__).parent
+    ssl_dir = test_dir / "ssl"
+    proj_dir = test_dir.parent
+    support_dir = proj_dir / "virtualenv_support"
+    local_pypi_app = pypiserver.app(root=str(support_dir))
+    local_pypi = pytest_localserver.http.WSGIServer(
+        host="localhost",
+        port=0,
+        application=local_pypi_app,
+        ssl_context=(str(ssl_dir / "server.crt"), str(ssl_dir / "server.key")),
+    )
+    local_pypi.start()
+    local_pypi_url = "https://localhost:{}/".format(local_pypi.server_address[1])
+    venvdir = tmpdir / "venv"
+    pip_log = tmpdir / "pip.log"
+    env_addition = {
+        "PIP_CERT": str(ssl_dir / "rootCA.pem"),
+        "PIP_INDEX_URL": local_pypi_url,
+        "PIP_LOG": str(pip_log),
+        "PIP_RETRIES": "0",
+    }
+    if six.PY2:
+        env_addition = {key.encode("utf-8"): value.encode("utf-8") for key, value in env_addition.items()}
+    env_backup = {}
+    for key, value in env_addition.items():
+        if key in os.environ:
+            env_backup[key] = os.environ[key]
+        os.environ[key] = value
+    try:
+        virtualenv.create_environment(str(venvdir), download=True)
+        with pip_log.open("rb") as f:
+            assert b"SSLError" not in f.read()
+    finally:
+        local_pypi.stop()
+        for key in env_addition.keys():
+            os.environ.pop(key)
+            if key in env_backup:
+                os.environ[key] = env_backup[key]
+
+
+def check_pypy_pre_import():
+    import sys
+
+    # These modules(module_name, optional) are taken from PyPy's site.py:
+    # https://bitbucket.org/pypy/pypy/src/d0187cf2f1b70ec4b60f10673ff081bdd91e9a17/lib-python/2.7/site.py#lines-532:539
+    modules = [
+        ("encodings", False),
+        ("exceptions", True),  # "exceptions" module does not exist in Python3
+        ("zipimport", True),
+    ]
+
+    for module, optional in modules:
+        if not optional or module in sys.builtin_module_names:
+            assert module in sys.modules, "missing {!r} in sys.modules".format(module)
+
+
+@pytest.mark.skipif("platform.python_implementation() != 'PyPy'")
+def test_pypy_pre_import(tmp_path):
+    """For PyPy, some built-in modules should be pre-imported because
+    some programs expect them to be in sys.modules on startup.
+    """
+    check_code = inspect.getsource(check_pypy_pre_import)
+    check_code = textwrap.dedent(check_code[check_code.index("\n") + 1 :])
+    if six.PY2:
+        check_code = check_code.decode()
+
+    check_prog = tmp_path / "check-pre-import.py"
+    check_prog.write_text(check_code)
+
+    ve_path = str(tmp_path / "venv")
+    virtualenv.create_environment(ve_path)
+
+    bin_dir = virtualenv.path_locations(ve_path)[-1]
+
+    try:
+        cmd = [
+            os.path.join(bin_dir, "{}{}".format(virtualenv.EXPECTED_EXE, ".exe" if virtualenv.IS_WIN else "")),
+            str(check_prog),
+        ]
+        subprocess.check_output(cmd, universal_newlines=True, stderr=subprocess.STDOUT)
+    except subprocess.CalledProcessError as exception:
+        assert not exception.returncode, exception.output
+
+
+@pytest.mark.skipif(not hasattr(os, "symlink"), reason="requires working symlink implementation")
+def test_create_environment_with_exec_prefix_pointing_to_prefix(tmpdir):
+    """Create virtual environment for Python with ``sys.exec_prefix`` pointing
+    to ``sys.prefix`` or ``sys.base_prefix`` or ``sys.real_prefix`` under a
+    different name
+    """
+    venvdir = str(tmpdir / "venv")
+    python_dir = tmpdir / "python"
+    python_dir.mkdir()
+    path_key = str("PATH")
+    old_path = os.environ[path_key]
+    if hasattr(sys, "real_prefix"):
+        os.environ[path_key] = os.pathsep.join(
+            p for p in os.environ[path_key].split(os.pathsep) if not p.startswith(sys.prefix)
+        )
+    python = virtualenv.resolve_interpreter(os.path.basename(sys.executable))
+    try:
+        subprocess.check_call([sys.executable, "-m", "virtualenv", "-p", python, venvdir])
+        home_dir, lib_dir, inc_dir, bin_dir = virtualenv.path_locations(venvdir)
+        assert not os.path.islink(os.path.join(lib_dir, "distutils"))
+    finally:
+        os.environ[path_key] = old_path
+
+
+@pytest.mark.skipif(not hasattr(sys, "real_prefix"), reason="requires running from inside virtualenv")
+def test_create_environment_from_virtual_environment(tmpdir):
+    """Create virtual environment using Python from another virtual environment
+    """
+    venvdir = str(tmpdir / "venv")
+    home_dir, lib_dir, inc_dir, bin_dir = virtualenv.path_locations(venvdir)
+    virtualenv.create_environment(venvdir)
+    assert not os.path.islink(os.path.join(lib_dir, "distutils"))
